#include <iostream>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <queue>
#include <cfloat> //Novo
#include "aig_parser.h"

//#define MIN(x,y) (((x)<(y))?(x):(y))
//#define MAX(x,y) (((x)<(y))?(y):(x))

using namespace std;

/*Attributes*/
/*13*/
long double * varNodesInFtree;
/*2*/
long double * andsInFtree;
long double * notsInFtree;
//long double * pnOccurences;
long double * pOccurences;
long double * nOccurences;
long double * avgNodeLeafPathLen;
long double * numNodeLeafPaths;
long double * stdDevNLPathLen;
long double *  minNLPathLen;
long double *  maxNLPathLen;
int overflow=0;
struct 
{
  long numVars;//1. Number of variables
  long double numConnTree; // Number of logical connectives in formula tree
  long double numConnDag; // Number of logical connectives in formula DAG
  long double treeSize; // Size of formula tree
  long double dagSize; // Size of formula DAG
  long double compressionFactor; // treeSize/dagSize
  long double compressionFactorRec; // dagSize/treeSize
  long double numVarsTree; // Number of variable nodes in formula tree
  long double connVarRatio; //Ratio of nodes representing connectives and nodes representing varibles ...
  long double varConnRatio; // ... and its reciprocal
  long double treeConnFreqAnd;
  long double treeConnFreqNot;
  //long double treeConnFreqOr;
  //long double treeConnFreqImp;
  //long double treeConnFreqEqv;
  long double treeConnEntropy;
  long double dagConnFreqAnd;
  long double dagConnFreqNot;
  //long double dagConnFreqOr;
  //long double dagConnFreqImp;
  //long double dagConnFreqEqv;
  long double dagConnEntropy;
  long double connCorrelation;
  //24-27
  long double varsTreeDistMean;
  long double varsTreeDistMax;
  long double varsTreeDistMin;
  long double varsTreeDistEntropy;
  //28-30
  long double polarityRatioMean;
  long double polarityRatioVariation;
  long double polarityRatioEntropy;
  //31-33
  long double orderedPolarityRatioMean;
  long double orderedPolarityRatioVariation;
  long double orderedPolarityRatioEntropy;
  // ~ 34 NLpaths
  long double nlPathMean;
  long double nlPathMin;
  long double nlPathMax;
  long double nlPathDeviation;
  long double nlPathVariation;
  //~ 35-39
  long double connArityMean;
  long double connArityMin;
  long double connArityMax;
  long double connArityVariation;
  long double connArityEntropy;
  //~ 40-44
  long double connInMean;
  long double connInMin;
  long double connInMax;
  long double connInVariation;
  long double connInEntropy;
  // ~45-47
  long double sfPolarityRatioMean;
  long double sfPolarityRatioVariation;
  long double sfPolarityRatioEntropy;
  long double sfOrderedPolarityRatioMean;
  long double sfOrderedPolarityRatioVariation;
  long double sfOrderedPolarityRatioEntropy;
  // ~ 48-49
  long double dagDepthVarNumRatio;
  long double VarNumdagDepthRatio;
  // ~ 50-51
  long double edgesNodesRatioDag;
  long double nodesEdgesRatioDag;


} attributes;

inline long double MIN(long double a, long double b)
{
  if(a<b)
    return a;
  else 
    return b;
}

AigParser::AigParser(char *inputFile)
{
    this->input=strdup(inputFile);
    this->aig=aiger_init();
    num_nots=0;
}

AigParser::~AigParser()
{
    free(this->input);
    aiger_reset(this->aig);
}

int AigParser::MakeTree()
{
    const char *error;
    error=aiger_open_and_read_from_file(aig, input);
    if(error)
    {
	cerr<<error<<endl;
	return AIG_PARSE_ERROR; //after this AigParser object must be destroyed and reinitialized
    }
    //allocate space for entire DAG
    long numElements=2*(aig->num_inputs+aig->num_ands);
    int numNodes=(int) numElements;
    if(numNodes!=numElements || (dag=(Node **) calloc((numNodes+2), sizeof(Node*)))==NULL)//+2 for indexing
    {
	std::cerr<<"Unable to allocate DAG\n";
	return AIG_ALLOC_ERROR;
    }
    
    for(unsigned int i=0;i<aig->num_inputs;i++)
    {
	dag[aig->inputs[i].lit]=new Variable(); 
	//cout<<"Var: "<<aig->inputs[i].lit<<endl;
//	dag[aig->inputs[i].lit+1]=new Negation(aig->inputs[i].lit);
//	dag[aig->inputs[i].lit]->addParent(aig->inputs[i].lit+1);
    }
    
    for(unsigned int i=0;i<aig->num_ands;i++)
    {
	if( (aig->ands[i].rhs0%2) && dag[aig->ands[i].rhs0]==NULL)
	{
	    dag[aig->ands[i].rhs0]=new Negation(aig->ands[i].rhs0-1);
	    dag[aig->ands[i].rhs0-1]->addParent(aig->ands[i].rhs0);
	    num_nots++;
	    //cout<<"Not: "<<aig->ands[i].rhs0<<endl;
	}
	if( (aig->ands[i].rhs1%2) && dag[aig->ands[i].rhs1]==NULL)
	{
	    dag[aig->ands[i].rhs1]=new Negation(aig->ands[i].rhs1-1);
	    dag[aig->ands[i].rhs1-1]->addParent(aig->ands[i].rhs1);
	    num_nots++;
	    //cout<<"Not: "<<aig->ands[i].rhs1<<endl;
	}    

	dag[aig->ands[i].lhs]=new Conjunction(aig->ands[i].rhs0,aig->ands[i].rhs1);
	//cout<<"And: "<<aig->ands[i].lhs<<" "<<aig->ands[i].rhs0<<" "<<aig->ands[i].rhs1<<endl;
	dag[aig->ands[i].rhs0]->addParent(aig->ands[i].lhs);
	dag[aig->ands[i].rhs1]->addParent(aig->ands[i].lhs);
//	dag[aig->ands[i].lhs+1]=new Negation(aig->ands[i].lhs);
//	dag[aig->ands[i].lhs]->addParent(aig->ands[i].lhs+1);
    }
//    for(int i=2;i<numNodes+2;i++)
    //   {
//	if(dag[i]->getType()==NOT && dag[i]->getNumParents()==0)
//	{
	    
    root=aig->outputs[0].lit;
    if(root!=0 && root!=1 && root%2)
    {
	dag[root]=new Negation(root-1);
	dag[root-1]->addParent(root);
    }
    return AIG_OK;
}

void AigParser::topDown()
{
  //int numInputs=aig->num_inputs;
  queue<int> q;
  //queue<short> v; 
  int cur;
  //int val;
  int child,parent;
  long double swap,delta, nStdDev, paths;
  q.push(root);
  //v.push(0);
  pOccurences[root]=1;
  nOccurences[root]=0;
  //numNodeLeafPaths[root]=1;
  avgNodeLeafPathLen[root]=0;
  stdDevNLPathLen[root]=0;
  minNLPathLen[root]=0;
  maxNLPathLen[root]=0;
  //cout<<"Size of long double:"<<sizeof(long double)<<endl;
  //cout<<"Min ld val: "<<LDBL_MIN<<endl<<"Max ld val: "<<LDBL_MAX<<endl;
  while(!q.empty())
    {
      cur=q.front();

      //val++;
      //Faza propagacije unazad
      swap=0;paths=0;
      
     
      
      for(int i=0;i<dag[cur]->getNumParents();i++)
	{
	  parent=dag[cur]->getParent(i);

	  delta=avgNodeLeafPathLen[cur]-avgNodeLeafPathLen[parent];
	  nStdDev=stdDevNLPathLen[parent]+ (delta-1)*(delta-1);
	  
	  /*cout<<"Cur: "<<cur<<endl<<"Parent: "<<parent<<endl;
	  cout<<"avgParent: "<<avgNodeLeafPathLen[parent]<<endl;
	  cout<<"avgCur: "<<avgNodeLeafPathLen[cur]<<endl;
	  cout<<"Delta: "<<delta<<endl;*/

 
	  //mozda moze sa manje operacija
	  //swap=paths/(paths + numNodeLeafPaths[parent]);
	  swap=paths/(paths + pOccurences[parent]+nOccurences[parent]);
	  //cout<<"Swap: "<<swap<<endl<<"PrevStdDev: "<<stdDevNLPathLen[cur]<<endl;
	  //cout<<"PathsSoFar: "<<paths<<endl<<"PathsFromParent: "<<pOccurences[parent]+nOccurences[parent]<<endl;
	  //cout<<"NoviStdDev: "<<(
	  stdDevNLPathLen[cur]=swap*stdDevNLPathLen[cur] + (1-swap)*nStdDev;//<<endl<<endl<<endl;
	  //paths+=numNodeLeafPaths[parent];
	  paths+= pOccurences[parent]+nOccurences[parent];
	}
      //Priprema za propagaciju unapred (ne valja da ide ispred koda za propagaciju unazad, kvari racun za listove?)
      if(dag[cur]->getType()==NOT)
	{
	  swap=pOccurences[cur];
	  pOccurences[cur]=nOccurences[cur];
	  nOccurences[cur]=swap;
	}
      /*if(dag[cur]->getType()!=VAR)
	{
	  avgNodeLeafPathLen[cur]+=1;
	  }*/
      // cout<<cur<<": "<<pOccurences[cur]<<"/"<<nOccurences[cur]<<endl<<"**********************************************"<<endl;
      //Faza propagacije unapred
      for(int i=0;i<dag[cur]->getNumChildren();i++)
	{
	  child=dag[cur]->getChild(i);

	  
	  //avgNodeLeafPathLen[child]=(avgNodeLeafPathLen[child]*numNodeLeafPaths[child] + avgNodeLeafPathLen[cur]*numNodeLeafPaths[cur])/(numNodeLeafPaths[child]+=numNodeLeafPaths[cur]);
	  //swap=numNodeLeafPaths[child]/(numNodeLeafPaths[child]+numNodeLeafPaths[cur]);
	  swap= (pOccurences[child]+nOccurences[child])/(pOccurences[child]+nOccurences[child] + pOccurences[cur]+nOccurences[cur]);
	  avgNodeLeafPathLen[child]=swap*avgNodeLeafPathLen[child] + (1-swap)*(avgNodeLeafPathLen[cur]+1);
	  //numNodeLeafPaths[child]+=numNodeLeafPaths[cur];

	  if(minNLPathLen[cur]+1<minNLPathLen[child] || 0==minNLPathLen[child])
	    minNLPathLen[child]=minNLPathLen[cur]+1;
	  if(maxNLPathLen[cur]+1>maxNLPathLen[child] || 0==maxNLPathLen[child])
	    maxNLPathLen[child]=maxNLPathLen[cur]+1;

	  pOccurences[child]+=pOccurences[cur];
	  nOccurences[child]+=nOccurences[cur];
	  //cout<<child<<": "<<pOccurences[child]<<"/"<<nOccurences[child]<<endl;
	  
	  dag[dag[cur]->getChild(i)]->increaseVisited();
	  if(dag[child]->getNumParents()==dag[child]->getVisited())
	    {
	      q.push(child);
	      dag[child]->resetVisited();		  
	    }


	}
      /*if(dag[cur]->getType()==VAR)
	{
	  cout<<cur<<": "<<pOccurences[cur]<<"/<<nOccurences[cur]<<endl;
	  }*/
      q.pop();
    }
}

int AigParser::calculateAttributes()
{
    //faza inicijacije, za svaki atribut koji racunamo pravimo globalni vektor

    if(aig->num_inputs==0)
      {
	constantsAttributes();
	return 0;
      }

    nOccurences=new long double[(int) 2*(aig->num_inputs+aig->num_ands)+2];
    pOccurences=new long double[(int) 2*(aig->num_inputs+aig->num_ands)+2];
    avgNodeLeafPathLen=new long double[(int) 2*(aig->num_inputs+aig->num_ands)+2];
    //numNodeLeafPaths=new long double[(int) 2*(aig->num_inputs+aig->num_ands)+2];
    stdDevNLPathLen=new long double[(int) 2*(aig->num_inputs+aig->num_ands)+2];
    minNLPathLen=new long double[(int) 2*(aig->num_inputs+aig->num_ands)+2];
    maxNLPathLen=new long double[(int) 2*(aig->num_inputs+aig->num_ands)+2];
    long double delta;
    /*for(unsigned int i=2;i<2*(aig->num_inputs+aig->num_ands)+2;i+=1)
      {
	stdDevNLPathLen[i]=0;
	}*/

     //COUNTING used VARS, ANDS, NOTS
    long numVariables=0;
    long numAnds=0;
    long numNots=0;
     for(unsigned long i=2;i<2*(aig->num_inputs+aig->num_ands)+2;i+=1)
      { 
	if(!dag[i] || (i!=root && dag[i]->getNumParents()==0))
	  continue;
	if(dag[i]->getType()==VAR)
	  numVariables++;
	if(dag[i]->getType()==NOT)
	  numNots++;
	if(dag[i]->getType()==AND)
	  numAnds++;
      }

    topDown();

    long double varNodesInFtree=0;
    long double andNodesInFtree=0;
    long double notNodesInFtree=0;

    for(unsigned int i=2;i<2*(aig->num_inputs+aig->num_ands)+2;i+=1)
      { 
	if(!dag[i]  || (i!=root && dag[i]->getNumParents()==0))
	  continue;

	if(dag[i]->getType()==VAR)
	  {
	    varNodesInFtree+=pOccurences[i]+nOccurences[i];
	    //cout<<i<<" "<<pOccurences[i]<<"/"<<nOccurences[i]<<endl;
	  }
	else if(dag[i]->getType()==AND)
	  andNodesInFtree+=pOccurences[i]+nOccurences[i];
	else if(dag[i]->getType()==NOT)
	  notNodesInFtree+=pOccurences[i]+nOccurences[i];
      }
    
    attributes.numVars=numVariables;

    attributes.numConnTree=andNodesInFtree+notNodesInFtree; //EXT
    if(isinf(attributes.numConnTree))//==INFINITY)
      {
	attributes.numConnTree=LDBL_MAX;
	overflow=1;
      }

    attributes.numConnDag=numAnds + numNots; //EXT

    attributes.numVarsTree=varNodesInFtree;
    if(isinf(attributes.numVarsTree))//==INFINITY)
      {
	attributes.numVarsTree=LDBL_MAX;
	overflow=1;
      }

    attributes.treeSize=attributes.numConnTree + attributes.numVarsTree;
    if(isinf(attributes.treeSize))//==INFINITY)
      {
	attributes.treeSize=LDBL_MAX;
	overflow=1;
      }

    attributes.dagSize=attributes.numConnDag + attributes.numVars;
    attributes.compressionFactor=attributes.treeSize/attributes.dagSize;
    attributes.compressionFactorRec=attributes.dagSize/attributes.treeSize;
    
    attributes.connVarRatio=attributes.numVarsTree/attributes.numConnTree;
    attributes.varConnRatio=attributes.numConnTree/attributes.numVarsTree;

    attributes.treeConnFreqAnd=andNodesInFtree/attributes.numConnTree;
    attributes.treeConnFreqNot=notNodesInFtree/attributes.numConnTree;
    if(isinf(attributes.treeConnFreqAnd) || isinf(attributes.treeConnFreqNot))
      {
	attributes.treeConnFreqAnd=0.5;
	attributes.treeConnFreqNot=0.5;	
      }

    //Entropiju treba prevesti u racunanje petljom
    if(attributes.treeConnFreqNot!=0)
      attributes.treeConnEntropy+= attributes.treeConnFreqNot*logl(attributes.treeConnFreqNot); 
    if(attributes.treeConnFreqAnd!=0)
      attributes.treeConnEntropy+= attributes.treeConnFreqAnd*logl(attributes.treeConnFreqAnd); //EXT    
    attributes.treeConnEntropy*=-1;
    attributes.dagConnFreqAnd=numAnds/attributes.numConnDag;
    attributes.dagConnFreqNot=numNots/attributes.numConnDag;

    if(attributes.dagConnFreqNot!=0)
      attributes.dagConnEntropy+= attributes.dagConnFreqNot*logl(attributes.dagConnFreqNot); 
    if(attributes.dagConnFreqAnd!=0)
      attributes.dagConnEntropy+= attributes.dagConnFreqAnd*logl(attributes.dagConnFreqAnd);
// attributes.dagConnEntropy= attributes.dagConnFreqNot*logl(attributes.dagConnFreqNot) + attributes.dagConnFreqAnd*logl(attributes.dagConnFreqAnd); //EXT
    attributes.dagConnEntropy*=-1;

    long double treeConnAvg=0.5*(attributes.treeConnFreqNot + attributes.treeConnFreqAnd);//EXT
    if(isinf(treeConnAvg))
      {
	treeConnAvg=LDBL_MAX;
	overflow=1;
      }
    long double treeConnStdDev=((attributes.treeConnFreqNot-treeConnAvg)*(attributes.treeConnFreqNot-treeConnAvg)+ (attributes.treeConnFreqAnd-treeConnAvg)*(attributes.treeConnFreqAnd-treeConnAvg));//EXT
    if(treeConnStdDev==0)
      treeConnStdDev=LDBL_MIN;
    long double dagConnAvg=0.5*(attributes.dagConnFreqNot+attributes.dagConnFreqAnd);//EXT
    long double dagConnStdDev=((attributes.dagConnFreqNot-dagConnAvg)*(attributes.dagConnFreqNot-dagConnAvg)+ (attributes.dagConnFreqAnd-dagConnAvg)*(attributes.dagConnFreqAnd-dagConnAvg));//EXT
    if(dagConnStdDev==0)
      dagConnStdDev=LDBL_MIN;
    attributes.connCorrelation=((attributes.dagConnFreqNot-dagConnAvg)*(attributes.treeConnFreqNot-treeConnAvg)+ (attributes.dagConnFreqAnd-dagConnAvg)*(attributes.treeConnFreqAnd-treeConnAvg))/sqrtl(treeConnStdDev*dagConnStdDev); //EXT
    // cout<<"tA: "<<treeConnAvg<<endl<<"tSD: "<<treeConnStdDev<<endl<<"dA: "<<dagConnAvg<<endl<<"dSD: "<<dagConnStdDev<<endl;
     //24-27
    attributes.varsTreeDistEntropy=0;
    for(int i=1;i<aig->num_inputs;i++)
      {
	if(dag[2*i] && dag[2*i]->getNumParents()!=0)
	  {
	    attributes.varsTreeDistMin=pOccurences[2*i]+nOccurences[2*i];
	    if(isinf(attributes.varsTreeDistMin))
	      attributes.varsTreeDistMin=LDBL_MAX;
	    // cout<<i<<": "<<attributes.varsTreeDistMin<<endl;
	    break;
	  }
      }
    attributes.varsTreeDistMax=attributes.varsTreeDistMin;
    attributes.varsTreeDistMean=0;
    
    attributes.nlPathMax=maxNLPathLen[2]; 
    attributes.nlPathMin=minNLPathLen[2]; 
    attributes.nlPathMean=0;
   
    long double swap, tmp;
    attributes.polarityRatioMean=0;
    attributes.orderedPolarityRatioMean=0;
    attributes.polarityRatioEntropy=0;
    attributes.orderedPolarityRatioEntropy=0;
    attributes.sfPolarityRatioMean=0;
    attributes.sfOrderedPolarityRatioMean=0;
    attributes.sfPolarityRatioEntropy=0;
    attributes.sfOrderedPolarityRatioEntropy=0;
    long double prStdDev=0;
    long double oprStdDev=0;
    long double sfprStdDev=0;
    long double sfoprStdDev=0;
    attributes.connArityEntropy=0;
    attributes.connArityMax=0;
    attributes.connArityMin=100;//dag[root]->getNumChildren();
    attributes.connArityMean=0;
    long double caStdDev=0;
    int nc;
    attributes.connInEntropy=0;
    attributes.connInMax=0;
    if(dag[root]->getNumChildren()>0)
      attributes.connInMin=dag[dag[root]->getChild(0)]->getNumParents();//dag[2]->getNumParents();
    else
      attributes.connInMin=0;
    attributes.connInMean=0;
    long double ciStdDev=0;
    int np;//potecijalno treba veci opseg
    long numEdges=0;
    //Prva for petlja za AVG
    long double * polarityRatioBins = new long double[100]; 
    long double * orderedPolarityRatioBins = new long double[100];
    long double * sfPolarityRatioBins = new long double[100];
    long double * sfOrderedPolarityRatioBins = new long double[100];
    long double * inboundBins = new long double[1501];
    long double * outboundBins = new long double[1501];

    
    for(int i=0; i<100;i++)
      {
	polarityRatioBins[i]=0;
	orderedPolarityRatioBins[i]=0;
	sfOrderedPolarityRatioBins[i]=0;
	sfPolarityRatioBins[i]=0;
      }
     for(int i=0; i<=1500;i++)
      {
	inboundBins[i]=0;
	outboundBins[i]=0;
      }

    //    long double * connInDagBins = new long double[];
     
    
     attributes.numVars=numVariables;
     for(unsigned long i=2;i<2*(aig->num_inputs+aig->num_ands)+2;i+=1)
      { 
	if(!dag[i] || (i!=root && dag[i]->getNumParents()==0))
	  continue;

	swap=MIN(pOccurences[i],nOccurences[i])/(pOccurences[i]+ nOccurences[i]);//MAX(pOccurences[i],nOccurences[i]);
	if(dag[i]->getType()==VAR)
	  {
	   
	    tmp=(pOccurences[i]+nOccurences[i])/attributes.numVarsTree;//varNodesInFtree[root];
	   
	    //cout<<i<<": "<<tmp<<endl;
	    attributes.varsTreeDistEntropy+=tmp*logl(tmp);
	    //cout<<i<<": "<<attributes.varsTreeDistEntropy<<endl;
	    
	    tmp=pOccurences[i]+nOccurences[i];
	    if(tmp<attributes.varsTreeDistMin)
	      {
		//cout<<"Tmp: "<<tmp<<endl;
		//cout<<"P: "<<pOccurences[i]<<endl;
		//cout<<"N: "<<nOccurences[i]<<endl;
		attributes.varsTreeDistMin=tmp;
		//cout<<(i/2)<<": "<<attributes.varsTreeDistMin<<endl;
	      }
	    if(tmp>attributes.varsTreeDistMax)
	      attributes.varsTreeDistMax=tmp;
     
	    attributes.polarityRatioMean+=pOccurences[i]/(pOccurences[i]+ nOccurences[i]);//nOccurences[i];
	    attributes.orderedPolarityRatioMean+=swap;//MIN(pOccurences[i],nOccurences[i])/MAX(pOccurences[i],nOccurences[i]);
	    
	    polarityRatioBins[(int)floorl(99.99*pOccurences[i]/(pOccurences[i]+ nOccurences[i]))]+=1;
	    //orderedPolarityRatioBins[(int)floorl(99.99*swap)]+=1;
	    orderedPolarityRatioBins[(int)(99.99*swap)/1]+=1;
	    if(minNLPathLen[i]<attributes.nlPathMin)
	      attributes.nlPathMin=minNLPathLen[i];
	    if(maxNLPathLen[i]>attributes.nlPathMax)
	      attributes.nlPathMax=maxNLPathLen[i];
	    //attributes.nlPathMean+=numNodeLeafPaths[i]*avgNodeLeafPathLen[i]/attributes.numVarsTree; //TODO: NodeLeafPaths replace with RootLeafPaths
	    attributes.nlPathMean+=(pOccurences[i]+nOccurences[i])*avgNodeLeafPathLen[i]/attributes.numVarsTree;
	  }
	else
	  {
	    nc=dag[i]->getNumChildren();
	    attributes.connArityMean+=nc;
	    if(attributes.connArityMax<nc)
	      attributes.connArityMax=nc;
	    if(attributes.connArityMin>nc )
	      attributes.connArityMin=nc;
	    
	    outboundBins[(int)MIN(nc,1500)]+=1;
	  }
	
 	swap=pOccurences[i]/(pOccurences[i]+ nOccurences[i]);
	attributes.sfPolarityRatioMean+=swap;//pOccurences[i]/(pOccurences[i]+ nOccurences[i]);
	sfPolarityRatioBins[(int)floorl(99.99*swap)]+=1;
	swap=MIN(pOccurences[i],nOccurences[i])/(pOccurences[i]+ nOccurences[i]);
	attributes.sfOrderedPolarityRatioMean+=swap;//MIN(pOccurences[i],nOccurences[i])/MAX(pOccurences[i],nOccurences[i]);
	sfOrderedPolarityRatioBins[(int)floorl(99.99*swap)]+=1;
	np=dag[i]->getNumParents();
	attributes.connInMean+=np;
	if(attributes.connInMax<np)
	  attributes.connInMax=np;
	if(attributes.connInMin>np && i!=root)
	  attributes.connInMin=np;    
	inboundBins[(int)MIN(np,1500)]+=1;
     
	numEdges+=dag[i]->getNumParents();
      }
   
    
    if(isinf(attributes.varsTreeDistMax))
      attributes.varsTreeDistMax=LDBL_MAX;
    if(overflow && !isnormal(attributes.varsTreeDistEntropy))
      attributes.varsTreeDistEntropy=0;
    else
      attributes.varsTreeDistEntropy*=-1;
    attributes.varsTreeDistMean=attributes.numVarsTree/attributes.numVars;
   
    attributes.polarityRatioMean/=attributes.numVars;
    attributes.orderedPolarityRatioMean/=attributes.numVars;
    
    attributes.sfPolarityRatioMean/=attributes.dagSize;
    attributes.sfOrderedPolarityRatioMean/=attributes.dagSize;
    
    attributes.connArityMean/=attributes.numConnDag;
    
    attributes.connInMean/=attributes.dagSize-1;
    
    
    //cout<<"BROJ PROMENLJIVIH: "<<numVariables<<endl;
    //Druga for petlja za STDDEV
    for(unsigned long i=2;i<2*(aig->num_inputs+aig->num_ands)+2;i+=1)
      {
	if(!dag[i]  || (i!=root && dag[i]->getNumParents()==0))
       continue;
	
	swap=MIN(pOccurences[i],nOccurences[i])/(pOccurences[i]+ nOccurences[i]);
	if(dag[i]->getType()==VAR)
	  {
	    prStdDev+=(pOccurences[i]/(pOccurences[i]+nOccurences[i]) - attributes.polarityRatioMean)*(pOccurences[i]/(pOccurences[i]+nOccurences[i]) - attributes.polarityRatioMean);
	    oprStdDev+=(swap-attributes.orderedPolarityRatioMean)*(swap-attributes.orderedPolarityRatioMean);
	    
	    delta=attributes.nlPathMean-avgNodeLeafPathLen[i];
	    attributes.nlPathDeviation+=(pOccurences[i]+nOccurences[i])*(stdDevNLPathLen[i]+delta*delta)/attributes.numVarsTree;
	  }
	else
	  {
	    nc=dag[i]->getNumChildren();
	    caStdDev+=(nc-attributes.connArityMean)*(nc-attributes.connArityMean);
	  }
	
	sfprStdDev+=(pOccurences[i]/(pOccurences[i]+nOccurences[i]) - attributes.sfPolarityRatioMean)*(pOccurences[i]/(pOccurences[i]+nOccurences[i]) - attributes.sfPolarityRatioMean);
	sfoprStdDev+=(swap-attributes.sfOrderedPolarityRatioMean)*(swap-attributes.sfOrderedPolarityRatioMean);      

    	np=dag[i]->getNumParents();
	if(i!=root)
	  ciStdDev+=(np-attributes.connInMean)*(np-attributes.connInMean);     
      }

    if(attributes.polarityRatioMean!=0)
      attributes.polarityRatioVariation=sqrtl(prStdDev/attributes.numVars)/attributes.polarityRatioMean;
    else
      attributes.polarityRatioVariation=0;

    if(attributes.orderedPolarityRatioMean!=0)
      attributes.orderedPolarityRatioVariation=sqrtl(oprStdDev/attributes.numVars)/attributes.orderedPolarityRatioMean;
    else
      attributes.orderedPolarityRatioVariation=0;

    if(attributes.sfPolarityRatioMean!=0)
      attributes.sfPolarityRatioVariation=sqrtl(sfprStdDev/attributes.dagSize)/attributes.sfPolarityRatioMean;
    else
      attributes.sfPolarityRatioMean=0;
    if(attributes.sfOrderedPolarityRatioMean!=0)
      attributes.sfOrderedPolarityRatioVariation=sqrtl(sfoprStdDev/attributes.dagSize)/attributes.sfOrderedPolarityRatioMean;
    else
      attributes.sfOrderedPolarityRatioVariation=0;

    attributes.connArityVariation=sqrtl(caStdDev/attributes.numConnDag)/attributes.connArityMean;

    attributes.connInVariation=sqrtl(ciStdDev/attributes.dagSize)/attributes.connInMean;
    
    attributes.dagDepthVarNumRatio=attributes.nlPathMax/attributes.numVars;
    attributes.VarNumdagDepthRatio=attributes.numVars/attributes.nlPathMax;
   
    attributes.nodesEdgesRatioDag=attributes.dagSize/numEdges;
    attributes.edgesNodesRatioDag=numEdges/attributes.dagSize;

    attributes.nlPathDeviation=sqrtl(attributes.nlPathDeviation);
    attributes.nlPathVariation=attributes.nlPathDeviation/attributes.nlPathMean;
    for(int i=0; i<100;i++)
      {
	if((swap=polarityRatioBins[i]/(long double)attributes.numVars)>0)
	  attributes.polarityRatioEntropy+=swap*logl(swap);

	if((swap=orderedPolarityRatioBins[i]/(long double)attributes.numVars)>0)
	  attributes.orderedPolarityRatioEntropy+=swap*logl(swap);
	if((swap=sfPolarityRatioBins[i]/(long double)attributes.dagSize)>0)
	  attributes.sfPolarityRatioEntropy+=swap*logl(swap);
	if((swap=sfOrderedPolarityRatioBins[i]/attributes.dagSize)>0)
	  attributes.sfOrderedPolarityRatioEntropy+=swap*logl(swap);
      }
    attributes.polarityRatioEntropy*=-1;
    attributes.orderedPolarityRatioEntropy*=-1;
    attributes.sfPolarityRatioEntropy*=-1;
    attributes.sfOrderedPolarityRatioEntropy*=-1;

    for(int i=0;i<=1500;i++)
      {
	if((swap=inboundBins[i]/attributes.dagSize)>0)
	  attributes.connInEntropy+=swap*logl(swap);
	if((swap=outboundBins[i]/attributes.numConnDag)>0)
	  attributes.connArityEntropy+=swap*logl(swap);
	//TODO outbound if nonAIG is used 
     }
    attributes.connArityEntropy*=-1;
    attributes.connInEntropy*=-1;

    if(overflow)
      {
	attributes.polarityRatioMean=0.5;
	attributes.polarityRatioVariation=1;
	//attributes.polarityRatioEntropy=0;
	
	attributes.orderedPolarityRatioMean=0.5;
	attributes.orderedPolarityRatioVariation=1;
	//attributes.orderedPolarityRatioEntropy=0;
	
	attributes.sfPolarityRatioMean=0.5;
	attributes.sfPolarityRatioVariation=1;
	//attributes.sfPolarityRatioEntropy=0;
	
	attributes.sfOrderedPolarityRatioMean=0.5;
	attributes.sfOrderedPolarityRatioVariation=1;
	//attributes.sfOrderedPolarityRatioEntropy=0;
	
	
	if(!isnormal(attributes.nlPathMean))
	  {
	    if(isfinite(attributes.nlPathMax) && isfinite(attributes.nlPathMin))
	      attributes.nlPathMean=0.5*(attributes.nlPathMax + attributes.nlPathMin);
	    else
	      {
		attributes.nlPathMean=LDBL_MAX;
		attributes.nlPathMin=LDBL_MAX;
		attributes.nlPathMax=LDBL_MAX;
		
	      }	
	    
	  }
	if(!isfinite(attributes.nlPathVariation))
	   attributes.nlPathVariation=1;
	
	
      }

    delete nOccurences;
    delete pOccurences;
    delete avgNodeLeafPathLen;
    //delete numNodeLeafPaths;
    delete stdDevNLPathLen;
    delete minNLPathLen;
    delete maxNLPathLen;

    delete polarityRatioBins;
    delete orderedPolarityRatioBins;
    delete sfPolarityRatioBins;
    delete sfOrderedPolarityRatioBins;
    delete inboundBins;
    delete outboundBins;



    return 0;
}

void AigParser::printAttributes()
{
    cout<<attributes.numVars;//1. Number of variables
    cout<<", "<<attributes.numConnTree;
    cout<<", "<<attributes.treeSize;
    cout<<", "<<attributes.dagSize;
    cout<<", "<<attributes.compressionFactor;
    cout<<", "<<attributes.compressionFactorRec;
    
    cout<<", "<<attributes.numVarsTree;//1. Number of variables
    cout<<", "<<attributes.connVarRatio;
    cout<<", "<<attributes.varConnRatio;
    cout<<", "<<attributes.treeConnFreqAnd;
    cout<<", "<<attributes.treeConnFreqNot;
    cout<<", "<<attributes.treeConnEntropy;
    cout<<", "<<attributes.dagConnFreqAnd;
    cout<<", "<<attributes.dagConnFreqNot;
    cout<<", "<<attributes.dagConnEntropy;
    cout<<", "<<attributes.connCorrelation;
    cout<<", "<<attributes.varsTreeDistMean;
    cout<<", "<<attributes.varsTreeDistMax;
    cout<<", "<<attributes.varsTreeDistMin;
    cout<<", "<<attributes.varsTreeDistEntropy;
    cout<<", "<<attributes.polarityRatioMean;
    cout<<", "<<attributes.polarityRatioVariation;
    cout<<", "<<attributes.polarityRatioEntropy;
    cout<<", "<<attributes.orderedPolarityRatioMean;
    cout<<", "<<attributes.orderedPolarityRatioVariation;
    cout<<", "<<attributes.orderedPolarityRatioEntropy;
    cout<<", "<<attributes.nlPathMean;
    cout<<", "<<attributes.nlPathMin;
    cout<<", "<<attributes.nlPathMax;
    cout<<", "<<attributes.nlPathVariation;
    
    cout<<", "<<attributes.connArityMean;
    cout<<", "<<attributes.connArityMin;
    cout<<", "<<attributes.connArityMax;
    cout<<", "<<attributes.connArityVariation;
    cout<<", "<<attributes.connArityEntropy;

    cout<<", "<<attributes.connInMean;
    cout<<", "<<attributes.connInMin;
    cout<<", "<<attributes.connInMax;
    cout<<", "<<attributes.connInVariation;
    cout<<", "<<attributes.connInEntropy;

    cout<<", "<<attributes.sfPolarityRatioMean;
    cout<<", "<<attributes.sfPolarityRatioVariation;
    cout<<", "<<attributes.sfPolarityRatioEntropy;
    cout<<", "<<attributes.sfOrderedPolarityRatioMean;
    cout<<", "<<attributes.sfOrderedPolarityRatioVariation;
    cout<<", "<<attributes.sfOrderedPolarityRatioEntropy;

    cout<<", "<<attributes.dagDepthVarNumRatio;
    cout<<", "<<attributes.VarNumdagDepthRatio;

    cout<<", "<<attributes.edgesNodesRatioDag;
    cout<<", "<<attributes.nodesEdgesRatioDag<<endl;
}

void AigParser::printAttributesVerbose()
{
    cout<<"numVars: "<<attributes.numVars<<endl;//1. Number of variables
    cout<<"numConnTree: "<<attributes.numConnTree<<endl;
    cout<<"treeSize: "<<attributes.treeSize<<endl;
    cout<<"dagSize: "<<attributes.dagSize<<endl;
    cout<<"compressionFactor: "<<attributes.compressionFactor<<endl;
    cout<<"compressionFactorRec: "<<attributes.compressionFactorRec<<endl;
    
    cout<<"numVarsTree: "<<attributes.numVarsTree<<endl;//1. Number of variables
    cout<<"dagConnVarRatio: "<<attributes.connVarRatio<<endl;
    cout<<"dagVarConnRatio: "<<attributes.varConnRatio<<endl;
    cout<<"treeConnFreqAnd: "<<attributes.treeConnFreqAnd<<endl;
    cout<<"treeConnFreqNot: "<<attributes.treeConnFreqNot<<endl;
    cout<<"treeConnEntropy: "<<attributes.treeConnEntropy<<endl;
    cout<<"dagConnFreqAnd: "<<attributes.dagConnFreqAnd<<endl;
    cout<<"dagConnFreqNot: "<<attributes.dagConnFreqNot<<endl;
    cout<<"dagConnEntropy: "<<attributes.dagConnEntropy<<endl;
    cout<<"connCorrelation: "<<attributes.connCorrelation<<endl;
    cout<<"varsTreeDistMean: "<<attributes.varsTreeDistMean<<endl;
    cout<<"varsTreeDistMax: "<<attributes.varsTreeDistMax<<endl;
    cout<<"varsTreeDistMin: "<<attributes.varsTreeDistMin<<endl;
    cout<<"varsTreeDistEntropy: "<<attributes.varsTreeDistEntropy<<endl;
    cout<<"polarityRatioMean: "<<attributes.polarityRatioMean<<endl;
    cout<<"polarityRatioVariation: "<<attributes.polarityRatioVariation<<endl;
    cout<<"polarityRatioEntropy: "<<attributes.polarityRatioEntropy<<endl;
    cout<<"orderedPolarityRatioMean: "<<attributes.orderedPolarityRatioMean<<endl;
    cout<<"orderedPolarityRatioVariation: "<<attributes.orderedPolarityRatioVariation<<endl;
    cout<<"orderedPolarityRatioEntropy: "<<attributes.orderedPolarityRatioEntropy<<endl;
    cout<<"nlPathMean: "<<attributes.nlPathMean<<endl;
    cout<<"nlPathMin: "<<attributes.nlPathMin<<endl;
    cout<<"nlPathMax: "<<attributes.nlPathMax<<endl;
    cout<<"nlPathVariation: "<<attributes.nlPathVariation<<endl;

    cout<<"connArityMean: "<<attributes.connArityMean<<endl;
    cout<<"connArityMin: "<<attributes.connArityMin<<endl;
    cout<<"connArityMax: "<<attributes.connArityMax<<endl;
    cout<<"connArityVariation: "<<attributes.connArityVariation<<endl;
    cout<<"connArityEntropy: "<<attributes.connArityEntropy<<endl;

    cout<<"connInMean: "<<attributes.connInMean<<endl;
    cout<<"connInMin: "<<attributes.connInMin<<endl;
    cout<<"connInMax: "<<attributes.connInMax<<endl;
    cout<<"connInVariation: "<<attributes.connInVariation<<endl;
    cout<<"connInEntropy: "<<attributes.connInEntropy<<endl;

    cout<<"sfPolarityRatioMean: "<<attributes.sfPolarityRatioMean<<endl;
    cout<<"sfPolarityRatioVariation: "<<attributes.sfPolarityRatioVariation<<endl;
    cout<<"sfPolarityRatioEntropy: "<<attributes.sfPolarityRatioEntropy<<endl;
    cout<<"sfOrderedPolarityRatioMean: "<<attributes.sfOrderedPolarityRatioMean<<endl;
    cout<<"sfOrderedPolarityRatioVariation: "<<attributes.sfOrderedPolarityRatioVariation<<endl;
    cout<<"sfOrderedPolarityRatioEntropy: "<<attributes.sfOrderedPolarityRatioEntropy<<endl;

    cout<<"dagDepthVarNumRatio: "<<attributes.dagDepthVarNumRatio<<endl;
    cout<<"VarNumdagDepthRatio: "<<attributes.VarNumdagDepthRatio<<endl;

    cout<<"edgesNodesRatioDag: "<<attributes.edgesNodesRatioDag<<endl;
    cout<<"nodesEdgesRatioDag: "<<attributes.nodesEdgesRatioDag<<endl;
    

}

void AigParser :: constantsAttributes()
{
attributes.numVars=0;
	attributes.numConnTree=0; 
	attributes.numConnDag=0; 
	attributes.treeSize=1; 
	attributes.dagSize=1;
	attributes.compressionFactor=1;
	attributes.compressionFactorRec=1;
	attributes.numVarsTree=0;
	attributes.connVarRatio=0; 
	attributes.varConnRatio=0;
	attributes.treeConnFreqAnd=0;
	attributes.treeConnFreqNot=0;
	attributes.treeConnEntropy=0;
	attributes.dagConnFreqAnd=0;
	attributes.dagConnFreqNot=0;
	attributes.dagConnEntropy=0;
	attributes.connCorrelation=0;
	attributes.varsTreeDistMean=0;
	attributes.varsTreeDistMax=0;
	attributes.varsTreeDistMin=0;
	attributes.varsTreeDistEntropy=0;
	attributes.polarityRatioMean=1;
	attributes.polarityRatioVariation=0;
	attributes.polarityRatioEntropy=0;
	
	attributes.orderedPolarityRatioMean=1;
	attributes.orderedPolarityRatioVariation=0;
	attributes.orderedPolarityRatioEntropy=0;

	attributes.nlPathMean=0;
	attributes.nlPathMin=0;
	attributes.nlPathMax=0;
	attributes.nlPathDeviation=0;
	attributes.nlPathVariation=0;
	
	attributes.connArityMean=0;
	attributes.connArityMin=0;
	attributes.connArityMax=0;
	attributes.connArityVariation=0;
	attributes.connArityEntropy=0;

	attributes.connInMean=0;
	attributes.connInMin=0;
	attributes.connInMax=0;
	attributes.connInVariation=0;
	attributes.connInEntropy=0;

	attributes.sfPolarityRatioMean=1;
	attributes.sfPolarityRatioVariation=0;
	attributes.sfPolarityRatioEntropy=0;
	attributes.sfOrderedPolarityRatioMean=1;
	attributes.sfOrderedPolarityRatioVariation=0;
	attributes.sfOrderedPolarityRatioEntropy=0;
	
	attributes.dagDepthVarNumRatio=0;
	attributes.VarNumdagDepthRatio=0;

	attributes.edgesNodesRatioDag=0;
	attributes.nodesEdgesRatioDag=0;
	
}
